/***********************************************************************
 * demo16x24.c - Arduino demo program for Holtek HT1632 LED driver chip,
 *            As implemented on the Sure Electronics DE-DP016 display board
 *            (16*24 dot matrix LED module.)
 * Nov, 2008 by Bill Westfield ("WestfW")
 *   Copyrighted and distributed under the terms of the Berkeley license
 *   (copy freely, but include this notice of original author.)
 *
 * Dec 2010, FlorinC - adapted for 3216 display
 ***********************************************************************/

#include <WProgram.h>
#include "ht1632.h"
#include <avr/pgmspace.h>
#include "font2.h"


#define X_MAX 64 //32
#define Y_MAX 16
#define CHIP_MAX 4*2 //Four HT1632Cs on one board
#define CLK_DELAY


// possible values for a pixel;
#define BLACK  0
#define GREEN  1
#define RED    2
#define ORANGE 3


#define plot(x,y,v)  ht1632_plot(x,y,v)
#define cls          ht1632_clear


// our own copy of the "video" memory; 64 bytes for each of the 4 screen quarters;
// each 64-element array maps 2 planes:
// indexes from 0 to 31 are allocated for green plane;
// indexes from 32 to 63 are allocated for red plane;
// when a bit is 1 in both planes, it is displayed as orange (green + red);
byte ht1632_shadowram[64][4*2] = {0};



/*
 * Set these constants to the values of the pins connected to the SureElectronics Module
 */
static const byte ht1632_data = 6;  // Data pin (pin 7 of display connector)
static const byte ht1632_wrclk = 7; // Write clock pin (pin 5 of display connector)
static const byte ht1632_cs = 8;    // Chip Select (pin 1 of display connnector)
static const byte ht1632_clk = 9; // clock pin (pin 2 of display connector)


#define DEMOTIME 10000  // 30 seconds max on each demo is enough.
#define DISPDELAY 40    // Each "display" lasts this long
#define LONGDELAY 1000  // This delay BETWEEN demos



//**************************************************************************************************
//Function Name: OutputCLK_Pulse
//Function Feature: enable CLK_74164 pin to output a clock pulse
//Input Argument: void
//Output Argument: void
//**************************************************************************************************
void OutputCLK_Pulse(void) //Output a clock pulse
{
  digitalWrite(ht1632_clk, HIGH);
  digitalWrite(ht1632_clk, LOW);
}


//**************************************************************************************************
//Function Name: OutputA_74164
//Function Feature: enable pin A of 74164 to output 0 or 1
//Input Argument: x: if x=1, 74164 outputs high. If x?1, 74164 outputs low.
//Output Argument: void
//**************************************************************************************************
void OutputA_74164(unsigned char x) //Input a digital level to 74164
{
  digitalWrite(ht1632_cs, (x==1 ? HIGH : LOW));
}


//**************************************************************************************************
//Function Name: ChipSelect
//Function Feature: enable HT1632C
//Input Argument: select: HT1632C to be selected
// If select=0, select none.
// If s<0, select all.
//Output Argument: void
//**************************************************************************************************
void ChipSelect(int select)
{
  unsigned char tmp = 0;
  if(select<0) //Enable all HT1632Cs
  {
    OutputA_74164(0);
    CLK_DELAY;
    for(tmp=0; tmp<CHIP_MAX; tmp++)
    {
      OutputCLK_Pulse();
    }
  }
  else if(select==0) //Disable all HT1632Cs
  {
    OutputA_74164(1);
    CLK_DELAY;
    for(tmp=0; tmp<CHIP_MAX; tmp++)
    {
      OutputCLK_Pulse();
    }
  }
  else
  {
    OutputA_74164(1);
    CLK_DELAY;
    for(tmp=0; tmp<CHIP_MAX; tmp++)
    {
      OutputCLK_Pulse();
    }
    OutputA_74164(0);
    CLK_DELAY;
    OutputCLK_Pulse();
    CLK_DELAY;
    OutputA_74164(1);
    CLK_DELAY;
    tmp = 1;
    for( ; tmp<select; tmp++)
    {
      OutputCLK_Pulse();
    }
  }
}


/*
 * ht1632_writebits
 * Write bits (up to 8) to h1632 on pins ht1632_data, ht1632_wrclk
 * Chip is assumed to already be chip-selected
 * Bits are shifted out from MSB to LSB, with the first bit sent
 * being (bits & firstbit), shifted till firsbit is zero.
 */
void ht1632_writebits (byte bits, byte firstbit)
{
  DEBUGPRINT(" ");
  while (firstbit) {
    DEBUGPRINT((bits&firstbit ? "1" : "0"));
    digitalWrite(ht1632_wrclk, LOW);
    if (bits & firstbit) {
      digitalWrite(ht1632_data, HIGH);
    } 
    else {
      digitalWrite(ht1632_data, LOW);
    }
    digitalWrite(ht1632_wrclk, HIGH);
    firstbit >>= 1;
  }
}


/*
 * ht1632_sendcmd
 * Send a command to the ht1632 chip.
 */
static void ht1632_sendcmd (byte chipNo, byte command)
{
  ChipSelect(chipNo);
  ht1632_writebits(HT1632_ID_CMD, 1<<2);  // send 3 bits of id: COMMMAND
  ht1632_writebits(command, 1<<7);  // send the actual command
  ht1632_writebits(0, 1); 	/* one extra dont-care bit in commands. */
  ChipSelect(0);
}


/*
 * ht1632_senddata
 * send a nibble (4 bits) of data to a particular memory location of the
 * ht1632.  The command has 3 bit ID, 7 bits of address, and 4 bits of data.
 *    Select 1 0 1 A6 A5 A4 A3 A2 A1 A0 D0 D1 D2 D3 Free
 * Note that the address is sent MSB first, while the data is sent LSB first!
 * This means that somewhere a bit reversal will have to be done to get
 * zero-based addressing of words and dots within words.
 */
static void ht1632_senddata (byte chipNo, byte address, byte data)
{
  ChipSelect(chipNo);
  ht1632_writebits(HT1632_ID_WR, 1<<2);  // send ID: WRITE to RAM
  ht1632_writebits(address, 1<<6); // Send address
  ht1632_writebits(data, 1<<3); // send 4 bits of data
  ChipSelect(0);
}


void ht1632_setup()
{
  pinMode(ht1632_cs, OUTPUT);
  digitalWrite(ht1632_cs, HIGH); 	/* unselect (active low) */
  pinMode(ht1632_wrclk, OUTPUT);
  pinMode(ht1632_data, OUTPUT);
  pinMode(ht1632_clk, OUTPUT);

  for (int j=1; j<5; j++)
  {
    ht1632_sendcmd(j, HT1632_CMD_SYSDIS);  // Disable system
    ht1632_sendcmd(j, HT1632_CMD_COMS00);
    ht1632_sendcmd(j, HT1632_CMD_MSTMD); 	/* Master Mode */
    ht1632_sendcmd(j, HT1632_CMD_RCCLK);  // HT1632C
    ht1632_sendcmd(j, HT1632_CMD_SYSON); 	/* System on */
    ht1632_sendcmd(j, HT1632_CMD_LEDON); 	/* LEDs on */
  }
  
  for (byte i=0; i<96; i++)
  {
    ht1632_senddata(1, i, 0);  // clear the display!
    ht1632_senddata(2, i, 0);  // clear the display!
    ht1632_senddata(3, i, 0);  // clear the display!
    ht1632_senddata(4, i, 0);  // clear the display!
  }
  delay(LONGDELAY);
}

//**************************************************************************************************
//Function Name: xyToIndex
//Function Feature: get the value of x,y
//Input Argument: x: X coordinate
//                y: Y coordinate
//Output Argument: address of xy
//**************************************************************************************************
byte xyToIndex(byte x, byte y)
{
  byte nChip, addr;

  if (x>=32) {
    nChip = 3 + x/16 + (y>7?2:0);
  } else {
    nChip = 1 + x/16 + (y>7?2:0);
  }

  x = x % 16;
  y = y % 8;
  addr = (x<<1) + (y>>2);

  return addr;
}

//**************************************************************************************************
//Function Name: calcBit
//Function Feature: calculate the bitval of y
//Input Argument: y: Y coordinate
//Output Argument: bitval
//**************************************************************************************************
#define calcBit(y) (8>>(y&3))

//**************************************************************************************************
//Function Name: get_pixel
//Function Feature: get the value of x,y
//Input Argument: x: X coordinate
//                y: Y coordinate
//Output Argument: color setted on x,y coordinates
//**************************************************************************************************
int get_pixel(byte x, byte y) {
  byte addr, bitval, nChip;

  if (x>=32) {
    nChip = 3 + x/16 + (y>7?2:0);
  } else {
    nChip = 1 + x/16 + (y>7?2:0);
  }

  addr = xyToIndex(x,y);
  bitval = calcBit(y);

  if ((ht1632_shadowram[addr][nChip-1] & bitval) && (ht1632_shadowram[addr+32][nChip-1] & bitval)) {
    return ORANGE;
  } else if (ht1632_shadowram[addr][nChip-1] & bitval) {
    return GREEN;
  } else if (ht1632_shadowram[addr+32][nChip-1] & bitval) {
    return RED;
  } else {
    return 0;
  }
}


/*
 * plot a point on the display, with the upper left hand corner
 * being (0,0), and the lower right hand corner being (31, 15);
 * parameter "color" could have one of the 4 values:
 * black (off), red, green or yellow;
 */
void ht1632_plot (byte x, byte y, byte color)
{
   byte nChip, addr, bitval;
  
  if (x<0 || x>X_MAX || y<0 || y>Y_MAX)
    return;
  
  if (color != BLACK && color != GREEN && color != RED && color != ORANGE)
    return;
  
  //byte nChip = 1 + x/16 + (y>7?2:0) ;
  
  if (x>=32) {
    nChip = 3 + x/16 + (y>7?2:0);
  } else {
    nChip = 1 + x/16 + (y>7?2:0);
  }
  
  // x = x % 16;
  // y = y % 8;
  
  
  addr = xyToIndex(x,y);
  bitval = calcBit(y);
  
  switch (color)
  {
    case BLACK:
      if (get_pixel(x,y) != BLACK) { // compare with memory to only set if pixel is other color
        // clear the bit in both planes;
        ht1632_shadowram[addr][nChip-1] &= ~bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
        addr = addr + 32;
        ht1632_shadowram[addr][nChip-1] &= ~bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
      }
      break;
    case GREEN:
      if (get_pixel(x,y) != GREEN) { // compare with memory to only set if pixel is other color
        // set the bit in the green plane and clear the bit in the red plane;
        ht1632_shadowram[addr][nChip-1] |= bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
        addr = addr + 32;
        ht1632_shadowram[addr][nChip-1] &= ~bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
      }
      break;
    case RED:
      if (get_pixel(x,y) != RED) { // compare with memory to only set if pixel is other color
        // clear the bit in green plane and set the bit in the red plane;
        ht1632_shadowram[addr][nChip-1] &= ~bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
        addr = addr + 32;
        ht1632_shadowram[addr][nChip-1] |= bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
      }
      break;
    case ORANGE:
      if (get_pixel(x,y) != ORANGE) { // compare with memory to only set if pixel is other color
        // set the bit in both the green and red planes;
        ht1632_shadowram[addr][nChip-1] |= bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
        addr = addr + 32;
        ht1632_shadowram[addr][nChip-1] |= bitval;
        ht1632_senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
      }
      break;
  }
}


void plot_test (int chipNo, byte addr, byte val)
{
   ht1632_shadowram[addr][0] = val;
   // Now copy the new memory value to the display
   ht1632_senddata(chipNo, addr, ht1632_shadowram[addr][0]);
}


/*
 * return the value of a pixel from the video memory (either BLACK, RED, GREEN, ORANGE);
 */
byte get_shadowram(byte x, byte y)
{
   byte nQuarter = x/16 + (y>7?2:0) ;
   x = x % 16;
   y = y % 8;
   byte addr = (x<<1) + (y>>2);
   byte bitval = 8>>(y&3);
   byte retVal = 0;
   byte val = (ht1632_shadowram[addr][nQuarter] & bitval) ? 1 : 0;
   val += (ht1632_shadowram[addr+32][nQuarter] & bitval) ? 2 : 0;
   return val;
}


/*
 * ht1632_clear
 * clear the display, and the shadow memory, and the snapshot
 * memory.  This uses the "write multiple words" capability of
 * the chipset by writing all 96 words of memory without raising
 * the chipselect signal.
 */
void ht1632_clear()
{
  char i;
  for (int i=1; i< 5; i++)
  {
    ChipSelect(-1);
    ht1632_writebits(HT1632_ID_WR, 1<<2);  // send ID: WRITE to RAM
    ht1632_writebits(0, 1<<6); // Send address
    for (i = 0; i < 96/2; i++) // Clear entire display
      ht1632_writebits(0, 1<<7); // send 8 bits of data
    ChipSelect(0);

    for (int j=0; j < 64; j++)
      ht1632_shadowram[j][i-1] = 0;
  }
}


void ht1632_putchar(byte x, byte y, char c, byte color=GREEN)
{
  byte dots;
  if (c >= 'A' && c <= 'Z' ||
    (c >= 'a' && c <= 'z') ) {
    c &= 0x1F;   // A-Z maps to 1-26
  } 
  else if (c >= '0' && c <= '9') {
    c = (c - '0') + 27;
  } 
  else if (c == ' ') {
    c = 0; // space
  }
  for (char col=0; col< 6; col++) {
    dots = pgm_read_byte_near(&myfont[c][col]);
    for (char row=0; row < 7; row++) {
      if (dots & (64>>row))   	     // only 7 rows.
        plot(x+col, y+row, color);
      else 
        plot(x+col, y+row, 0);
    }
  }
}


/*
 * snapshot_shadowram; used in Game of Life;
 * Copy the shadow ram into the snapshot ram (the upper bits)
 * This gives us a separate copy so we can plot new data while
 * still having a copy of the old data.  snapshotram is NOT
 * updated by the plot functions (except "clear")
 */
void snapshot_shadowram()
{
  for (int nQuarter=0; nQuarter<4; nQuarter++)
  {
    for (byte i=0; i<64; i++)
    {
      // copy the video bits (lower 4) in the upper 4;
      byte val = ht1632_shadowram[i][nQuarter];
      val = (val & 0x0F) + (val << 4);
      ht1632_shadowram[i][nQuarter] = val;
    }
  }
}


/*
 * get_snapshotram
 * get a pixel value from the snapshot ram instead of the actual video memory;
 * return BLACK, GREEN, RED or ORANGE;
 */
byte get_snapshotram(byte x, byte y)
{
   byte nQuarter = x/16 + (y>7?2:0);  // 0..3;
   x = x % 16;
   y = y % 8;
   byte addr = (x<<1) + (y>>2);
   byte bitval = 8>>(y&3);  // compute the required bit;
   
   byte greenByte = ht1632_shadowram[addr]   [nQuarter] >> 4;
   byte redByte   = ht1632_shadowram[addr+32][nQuarter] >> 4;

   byte retVal = 0;
   byte val = (greenByte & bitval) ? 1 : 0;
   val += (redByte & bitval) ? 2 : 0;
   return val;
}


/*
 * demo_life
 * Run the "life" game for a while, demonstrating the
 * ability of the AVR to update every pixle of the display
 * after having done some computation to figure out the new
 * value.  Also demonstrates the use of the snapshot ram.
 */
void demo_life ()
{
  byte x,y, neighbors, newval;
/*
  plot(10,9,1);  // Plant an "acorn"; a simple pattern that
  plot(12,10,1); //  grows for quite a while..
  plot(9,11,1);
  plot(10,8,1);
  plot(10,12,1);
  plot(10,13,1);
  plot(11,10,1);
  plot(11,11,1);
  plot(13,11,1);
  plot(14,11,1);
  plot(15,11,1);
*/
  delay(LONGDELAY);   // Play life
  snapshot_shadowram();

  for (int i=0; i < (DEMOTIME/DISPDELAY)/4; i++) {
    for (x=0; x < X_MAX; x++) {
      for (y=0; y < Y_MAX; y++) {
        neighbors = get_snapshotram(x, y+1) +
          get_snapshotram(x, y-1) +
          get_snapshotram(x+1, y) +
          get_snapshotram(x+1, y+1) +
          get_snapshotram(x+1, y-1) +
          get_snapshotram(x-1, y) +
          get_snapshotram(x-1, y+1) +
          get_snapshotram(x-1, y-1);

        switch (neighbors) {
        case 0:
        case 1:
          newval = 0;   // death by loneliness
          break;
        case 2:
          newval = 1;//get_snapshotram(x,y);
          break;  // remains the same
        case 3:
          newval = 2;
          break;
        default:
          newval = 0;  // death by overcrowding
          break;
        }
       
        plot(x,y, newval);
      }
    }
    snapshot_shadowram();
    delay(DISPDELAY);
  }
  delay(1000);
}


/*
 * Draw a line between two points using the bresenham algorithm.
 * This particular bit of code is copied nearly verbatim from
 *   http://www.gamedev.net/reference/articles/article1275.asp
 * I don't like it much (too many local variables!), but it works,
 * and is fully explained at the site, so...
 */
void bres_line(byte x1, byte y1,   byte x2, byte y2,  byte val )
{
  char deltax = abs(x2 - x1);        // The difference between the x's
  char deltay = abs(y2 - y1);        // The difference between the y's
  char x = x1;                       // Start x off at the first pixel
  char y = y1;                       // Start y off at the first pixel
  char xinc1, xinc2, yinc1, yinc2, den, num, numadd, numpixels, curpixel;

  if (x2 >= x1) {                // The x-values are increasing
    xinc1 = 1;
    xinc2 = 1;
  }  
  else {                          // The x-values are decreasing
    xinc1 = -1;
    xinc2 = -1;
  }

  if (y2 >= y1)                 // The y-values are increasing
  {
    yinc1 = 1;
    yinc2 = 1;
  }
  else                          // The y-values are decreasing
  {
    yinc1 = -1;
    yinc2 = -1;
  }

  if (deltax >= deltay)         // There is at least one x-value for every y-value
  {
    xinc1 = 0;                  // Don't change the x when numerator >= denominator
    yinc2 = 0;                  // Don't change the y for every iteration
    den = deltax;
    num = deltax / 2;
    numadd = deltay;
    numpixels = deltax;         // There are more x-values than y-values
  }
  else                          // There is at least one y-value for every x-value
  {
    xinc2 = 0;                  // Don't change the x for every iteration
    yinc1 = 0;                  // Don't change the y when numerator >= denominator
    den = deltay;
    num = deltay / 2;
    numadd = deltax;
    numpixels = deltay;         // There are more y-values than x-values
  }

  for (curpixel = 0; curpixel <= numpixels; curpixel++)
  {
    plot(x, y, val);             // Draw the current pixel
    num += numadd;              // Increase the numerator by the top of the fraction
    if (num >= den)             // Check if numerator >= denominator
    {
      num -= den;               // Calculate the new numerator value
      x += xinc1;               // Change the x as appropriate
      y += yinc1;               // Change the y as appropriate
    }
    x += xinc2;                 // Change the x as appropriate
    y += yinc2;                 // Change the y as appropriate
  }
}


/*
 * demo_chars
 * draw some characters on the display and scroll them vertically.
 */
void demo_chars ()
{
  for (int y=0; y <= Y_MAX-7; y++)
  {
    ht1632_putchar(0, y, 'A', random(BLACK, ORANGE+1));
    ht1632_putchar(6, y, 'B', random(BLACK, ORANGE+1));
    ht1632_putchar(12, y, '1', random(BLACK, ORANGE+1));
    ht1632_putchar(18, y, '3', random(BLACK, ORANGE+1));
    ht1632_putchar(24, y, 'm', random(BLACK, ORANGE+1));
    if (y>0)
      bres_line(0, y-1, 31, y-1, BLACK);
    delay(DISPDELAY);
  }
}


/***********************************************************************
 * traditional Arduino sketch functions: setup and loop.
 ***********************************************************************/

void setup ()  // flow chart from page 17 of datasheet
{
  ht1632_setup();  randomSeed(analogRead(0));
  
  Serial.begin(9600);

  cls();
}


void loop ()
{
  
ht1632_putchar(0, 0, 'A', GREEN);
ht1632_putchar(6, 0, 'B', RED);
ht1632_putchar(12, 0, 'C', ORANGE);
ht1632_putchar(18, 0, 'D', GREEN);
ht1632_putchar(24, 0, 'E', RED);
  
ht1632_putchar(30, 0, 'F', ORANGE);
ht1632_putchar(36, 0, 'G', GREEN);
ht1632_putchar(42, 0, 'H', RED);
ht1632_putchar(48, 0, 'I', ORANGE);
ht1632_putchar(54, 0, 'J', GREEN);
ht1632_putchar(60, 0, 'K', RED);

ht1632_putchar(0, 8, 'L', ORANGE);
ht1632_putchar(6, 8, 'M', GREEN);
ht1632_putchar(12, 8, 'N', RED);
ht1632_putchar(18, 8, 'O', ORANGE);
ht1632_putchar(24, 8, 'P', GREEN);
  
ht1632_putchar(30, 8, 'Q', RED);
ht1632_putchar(36, 8, 'R', ORANGE);
ht1632_putchar(42, 8, 'S', GREEN);
ht1632_putchar(48, 8, 'T', RED);
ht1632_putchar(54, 8, 'U', ORANGE);
ht1632_putchar(60, 8, 'V', GREEN);
  
// demo_chars();
// demo_life();
// demo_filler();
// demo_random_walk();
}


/*
 * demo_random_walk()
 * have a single LED walk all over the display, randomly.
 */
void demo_random_walk ()
{
  char x,y, dx, dy;
  dx = dy = 1;
  byte change;

  cls();
  for (int i=0; i < (DEMOTIME/DISPDELAY);  i++) {
    plot(x,y,1);  // draw a point
    delay(DISPDELAY); // wait a bit.
    change = random(1,32);
    /*
     * figure out where to go next.  This code is a bit
     * random in more senses than one, but it seems to
     * have results that are more or less what I had in
     * mind for this portion of the demo.
     */
    if (change < 9) {
      // do nothing
    } 
    else if (change == 10) {
      dx = -dx;
    } 
    else if (change == 11) {
      dy = -dy;
    } 
    else if (change == 12) {
      dx = dy = 0;
    } 
    else if (change == 13) {
      dx = dy = 1;
    } 
    else if (change == 14) {
      dx = dy = -1;
    }

    plot(x,y,0); // erase the point

    x = x + dx;
    y = y+ dy;
    if (x > X_MAX) {
      dx = -1;
      x = X_MAX;
    } 
    else if (x < 0) {
      x = 0;
      dx = 1;
    }
    if (y >  Y_MAX) {
      y = Y_MAX;
      dy = -1;
    } 
    else if (y < 0) {
      y = 0;
      dy = 1;
    }
  }
}


void demo_filler()
{
  Serial.println("fill with green dots");
  
  // green dots;
  for (int x = 0; x<32; x++)
  {
    for (int y = 0; y<16; y++)
    {
      ht1632_plot(x, y, GREEN);
      delay(20);
    }
  }

  Serial.println("fill with red dots");

  // red dots;
  for (int x = 0; x<32; x++)
  {
    for (int y = 0; y<16; y++)
    {
      ht1632_plot(x, y, RED);
      delay(20);
    }
  }

  Serial.println("fill with orange dots");

  // orange dots;
  for (int x = 0; x<32; x++)
  {
    for (int y = 0; y<16; y++)
    {
      ht1632_plot(x, y, ORANGE);
      delay(20);
    }
  }

  Serial.println("fill with black dots");

  // black dots;
  for (int x = 0; x<32; x++)
  {
    for (int y = 0; y<16; y++)
    {
      ht1632_plot(x, y, BLACK);
      delay(20);
    }
  }
}

/*
void loop ()
{
 for (int k=1; k<5; k++)
 {
   for (byte addr=0; addr<64; addr++)
   {
     for (int val=0; val<16; val++)
     {
       plot_test (k, addr, val);
       delay(20);
     }
   }
 }
  cls();
}
*/
